#include <cstring>
#include "StdConc.h"
#include "../common/util_classes.h"
#include "../common/DwdsThesaurus.h"
#include "ConcIndexator.h"

const int DefaultKwicContextSize = 4;

void CConcIndexator::InitDefaultOptions()
{
   	m_Path = "#empty_path";
	m_bUseParagraphTagToDivide = false;
	m_bEmptyLineIsSentenceDelim = true;
	m_bUseIndention = true;
	m_bIndexMorphPatterns = false;
	m_bUseDwdsThesaurus = false;
	m_UserMaxTokenCountInOnePeriod = 0;
	m_bUserMaxTokenCountInOnePeriod = false;
	m_bIndexPunctuation = true;
	m_bResumeOnIndexErrors = false;
	m_bQueryOnlyFiles = false;
	m_bDisableDefaultQueryLexicalExpansion = false;
	m_bUseDwdsThesaurus = false;
	m_bIndexChunks = false;
	m_bCaseSensitive = true;
	m_bOutputBibliographyOfHits = false;
	m_bDwdsCorpusInterface = false;
	m_bGutenbergInterface = false;
	m_bNoContextOperator = false;
	m_IndexType = DWDS_Index;
	m_bArchiveIndex = false;

	m_Language = morphUnknown;
	m_bShowNumberOfRelevantDocuments = false;
	m_HtmlHighlighting.m_FirstOpener = "<STRONG><FONT COLOR=red>";
	m_HtmlHighlighting.m_FirstCloser = "</FONT></STRONG>";
	m_HtmlHighlighting.m_RestOpener = "<STRONG><FONT COLOR=red>";
	m_HtmlHighlighting.m_RestCloser = "</FONT></STRONG>";
	m_TextHighlighting.m_FirstOpener = "&&";
	m_TextHighlighting.m_FirstCloser = "&&";
	m_TextHighlighting.m_RestOpener = "_&";
	m_TextHighlighting.m_RestCloser = "&_";
	m_LeftKwicContextSize = DefaultKwicContextSize;
	m_RightKwicContextSize = DefaultKwicContextSize;
	m_NumberOfKwicLinesInSnippets = 10;
	m_TfIdfRank = 0.3;
	m_NearRank = 0.3;
	m_PositionRank = 0.3;
	m_InterpDelimiter = "#";

}

const char* CConcIndexator::GetIndexTypeStr () const
{
	switch (m_IndexType)
	{
		case DWDS_Index:  return "DWDS_Index";
		case MorphXML_Index:  return "MorphXML_Index";
		case Free_Index:  return "Free_Index";
	};
	assert (false);
	return "unknown";
};

bool  CConcIndexator::ReadIndexTypeFromStr (const string& s)
{
	if (s == "DWDS_Index") 
		m_IndexType =  DWDS_Index;
	else
	if (s == "MorphXML_Index") 
		m_IndexType =  MorphXML_Index;
	else
	if (s == "Free_Index") 
		m_IndexType =  Free_Index;
	else
		return false;

	return true;
};



bool CConcIndexator::LoadOptionsFromString(string Options)
{
	StringTokenizer lines (Options.c_str(), "\n\r");
	MorphLanguageEnum langua;
	string IndicesStr, HitTypesStr, FreeBiblDescr, TextAreasDescr;
	m_Language = morphUnknown;
	m_IndicesToShow.resize(1,0);
    vector<string> BigramDescr;
	

	while (lines())
	{
		char s[1024];
		strcpy(s, lines.val());
		
		StringTokenizer fields(s, " \t\r\n");
		if (!fields()) continue;
		string q = fields.val();
		string q_lower = q;
		RmlMakeLower(q_lower, morphEnglish);
		
		if (q == "#include")
		{
			q = fields.get_rest();
			Trim(q);
			if ((q.length() < 2) || (q[0] != '\"') || (q[q.length() -1] != '\"'))
			{
				fprintf (stderr, "bad syntax in opt-file : %s!\n", s);
				return false;
			};
			string AddOptions;
			string IncFileName = q.substr(1,q.length()-2);
			if (!LoadFileToString(IncFileName , AddOptions))
			{
				fprintf (stderr, "cannot load file \"%s\"!\n", IncFileName.c_str());
				return false;
			};
			Options = AddOptions+"\n"+lines.get_rest();
			lines.reinitialize(Options.c_str(), "\n\r");
		}
		else
		if (q == "LocalPathPrefix")
		{
			q = fields.next_token();
			if (q.empty() )
			{
				fprintf (stderr, "bad syntax in opt-file : %s!\n", s);
				return false;
			};
			m_LocalPathPrefix = q;
		}
		else
			if (q == "InternetPathPrefix")
			{
				q = fields.next_token();
				if (q.empty() )
				{
					fprintf (stderr, "bad syntax in opt-file : %s!\n", s);
					return false;
				};

				m_InternetPathPrefix = q;
			}
		else
			if (GetLanguageByString(q, langua))
			{
				fprintf (stderr,"%s\n", GetStringByLanguage (langua).c_str());
				m_Language = langua;
			}
		else
			if (q == "CaseInsensitive")
			{
				m_bCaseSensitive = false;
				fprintf (stderr,"CaseInsensitive\n");
			}
		else
			if (q == "HtmlHighlighting")
			{
				if (!m_HtmlHighlighting.ReadFromString(fields.get_rest()))
				{
					fprintf (stderr,"Cannot parse HtmlHighlighting field\n");
					return false;
				};
				fprintf (stderr,"HtmlHighlighting is user-defined\n");
			}
		else
			if (q == "TextHighlighting")
			{
				if (!m_TextHighlighting.ReadFromString(fields.get_rest()))
				{
					fprintf (stderr,"Cannot parse TextHighlighting field\n");
					return false;
				};
				fprintf (stderr,"TextHighlighting is user-defined\n");
			}
		else
			if (q == "ShowNumberOfRelevantDocuments")
			{
				m_bShowNumberOfRelevantDocuments = true;
				fprintf (stderr,"ShowNumberOfRelevantDocuments\n");
			}
		else
			if (q == FreeBiblAttribOptionFieldName)
			{
				FreeBiblDescr += string(s) + "\n";
			}
		else
			if (q_lower == TextAreaOptionFieldName)
			{
				TextAreasDescr += string(s) + "\n";
			}
		else
			if (q == "UseParagraphTagToDivide")
			{
				m_bUseParagraphTagToDivide = true;
			}
		else
		if (q == "QueryOnlyFiles")
		{
				m_bQueryOnlyFiles = true;
				fprintf (stderr,"QueryOnlyFiles\n");
		}
		else
			if (q == "NoContextOperator" )
			{
				m_bNoContextOperator = true;
			}
			
		else
			if (q == "EmptyLineIsNotSentenceDelim")
			{
				m_bEmptyLineIsSentenceDelim = false;
			}
		else
			if (q == "DontUseIndention")
			{
				m_bUseIndention = false;
			}
		else
			if (q == "UseDwdsThesaurus")
			{
				m_bUseDwdsThesaurus = true;
				fprintf (stderr,"UseDwdsThesaurus\n");
			}
		else
			if (q == "MaxRegExpExpansionSize")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_MaxRegExpExpansionSize = atoi(q.c_str());
					fprintf (stderr,"MaxRegExpExpansionSize = %i\n", m_MaxRegExpExpansionSize);
				}
			}
		else
			if (q == "RightKwicContextSize")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_RightKwicContextSize = atoi(q.c_str());
					fprintf (stderr,"RightKwicContextSize = %i\n", m_RightKwicContextSize);
				}
				else
					return false;
			}
		else
			if (q == "LeftKwicContextSize")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_LeftKwicContextSize = atoi(q.c_str());
					fprintf (stderr,"LeftKwicContextSize = %i\n", m_LeftKwicContextSize);
				}
				else
					return false;
			}
		else
			if (q == "NumberOfKwicLinesInSnippets")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_NumberOfKwicLinesInSnippets = atoi(q.c_str());
					fprintf (stderr,"NumberOfKwicLinesInSnippets = %i\n", m_NumberOfKwicLinesInSnippets);
				}
				else
					return false;
			}
		else
			if (q == "TfIdfRank")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_TfIdfRank = atof(q.c_str());
					fprintf (stderr,"TfIdfRank = %f\n", m_TfIdfRank);
				}
				else
					return false;
			}
		else
			if (q == "PositionRank")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_PositionRank = atof(q.c_str());
					fprintf (stderr,"PositionRank = %f\n", m_PositionRank);
				}
				else
					return false;
			}
		else
			if (q == "InterpDelimiter")
			{
				q = fields.next_token();
				if (q.empty())  
				{
					fprintf (stderr,"InterpDelimiter is not set!\n");
					return false;
				}
				fprintf (stderr,"InterpDelimiter = %s\n", m_InterpDelimiter.c_str());
				m_InterpDelimiter = q;
			}
		else
			if (q == "IndicesToShow")
			{
				q = fields.next_token();
				if (q.empty())  return false;
				m_IndicesToShow.clear();
				while (!q.empty())
				{
					if (q == "0") 
					{
						fprintf (stderr,"Error! IndicesToShow must start from 1\n");
						return false;
					}
					int a = atoi(q.c_str());
					if (!a)
					{
						fprintf (stderr,"Error! Bad format of IndicesToShow, it must be a blank separated list of integers (started from 1!)!\n");
						return false;
					}
					for (size_t j=0; j<q.length(); j++)
						if (!isdigit((BYTE)q[j]))
						{
							fprintf (stderr,"Error! Bad format of IndicesToShow, it must be a blank separated list of integers (started from 1!)!\n");
							return false;
						}

					m_IndicesToShow.push_back(a-1);
					q = fields.next_token();
				}
				fprintf (stderr,"IndicesToShow is initialized with %i values\n", m_IndicesToShow.size());
			}
		else
			if (q == "NearRank")
			{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_NearRank = atof(q.c_str());
					fprintf (stderr,"NearRank = %f\n", m_NearRank);
				}
				else
					return false;
			}
			
		else
			if (q == "IndexType")
			{
				q = fields.next_token();
				if (!ReadIndexTypeFromStr(q)) 
				{
					fprintf (stderr,"Unknown value for IndexType\n");
					return false;
				}
				fprintf (stderr,"IndexType %s\n", GetIndexTypeStr());
			}
		else
			if (q == "GutenbergInterface")
			{
				m_bGutenbergInterface = true;
				fprintf (stderr,"GutenbergInterface\n");
			}
		else
			if (q == "DwdsCorpusInterface")
			{
				m_bDwdsCorpusInterface = true;
				fprintf (stderr,"DwdsCorpusInterface\n");
			}
		else
			if (q == "Bigrams")
			{
                BigramDescr.push_back( fields.get_rest() );
			}
		else
		if (q == "Indices")
		{
			IndicesStr = fields.get_rest();
		}
		else
		if (q == "HitBorders")
		{
			HitTypesStr = fields.get_rest();
		}
		else
		if (q == "IndexPunctuation")
		{
				m_bIndexPunctuation = true;
				fprintf (stderr,"IndexPunctuation\n");
		}
		else
		if (q == "ResumeOnIndexErrors")
		{
				m_bResumeOnIndexErrors = true;
				fprintf (stderr,"ResumeOnIndexErrors\n");
		}
		else
		if (q == "IndexMorphPatterns")
		{
				m_bIndexMorphPatterns = true;
				fprintf (stderr,"IndexMorphPatterns\n");
		}
		else
		if (q == "DisableDefaultQueryLexicalExpansion")
		{
				m_bDisableDefaultQueryLexicalExpansion = true;
				fprintf (stderr,"DisableDefaultQueryLexicalExpansion\n");
		}
		else
		if (q == "ArchiveIndex")
		{
				m_bArchiveIndex = true;
				fprintf (stderr,"ArchiveIndex\n");
		}
		else
		if (q == "OutputBibliographyOfHits")
		{
				m_bOutputBibliographyOfHits = true;
				fprintf (stderr,"OutputBibliographyOfHits\n");
		}
		else
		if (q == "IndexChunks")
		{
				m_bIndexChunks = true;
				fprintf (stderr,"IndexChunks\n");
		}
		else
		if (q == "UserMaxTokenCountInOnePeriod")
		{
				q = fields.next_token();
				if (!q.empty()) 
				{
					m_bUserMaxTokenCountInOnePeriod = true;
					m_UserMaxTokenCountInOnePeriod = atoi(q.c_str());
					

					fprintf (stderr,"UserMaxTokenCountInOnePeriod = %i\n", m_UserMaxTokenCountInOnePeriod);
					if (m_UserMaxTokenCountInOnePeriod == 0)
					{
						fprintf (stderr,"It cannot be less than 65000!\n");
						return false;
					};
				};
				
		}
		else
		{
			fprintf (stderr, "bad syntax in opt-file : %s!\n", q.c_str());
			return false;
		};
	};


	if (m_Language == morphUnknown)
	{
		ErrorMessage ("There is no language definition in the options file");
		return false;
	};

	RmlPcreMakeTables(m_PcreCharacterTables, m_Language);

	
	
	const char* ArchiveType = (m_bArchiveIndex) ? "archive" : "normal";
	
	switch (m_IndexType) 
	{
		case MorphXML_Index: 	IndicesStr = Format("[Token w %s];[MorphPattern m %s];[Lemma l %s]", 
									ArchiveType,
									ArchiveType,
									ArchiveType);
								break;

		case DWDS_Index:	IndicesStr = Format("[Token w %s];",ArchiveType);
							if (m_bIndexMorphPatterns)
								IndicesStr += Format("[MorphPattern m %s];",ArchiveType);
							if (m_bUseDwdsThesaurus)
								IndicesStr += Format("[Thes t %s];",ArchiveType);
							break;

		case Free_Index:	if ( IndicesStr.empty() )
							{
								fprintf (stderr, "Error! Cannot find an index definition in the options file.\n");
								return false;
							};
							break;
	};

	if (!RegisterStringIndices(IndicesStr))
	{
		fprintf (stderr," Cannot parse index defintion: %s\n", IndicesStr.c_str());
		return false;
	};




	if (m_bIndexChunks)
		if (!RegisterChunkIndex())
		{
			fprintf (stderr, "cannot register predefined chunk index!\n");
			return false;
		};

	fprintf (stderr," indices = %s\n", GetIndicesString().c_str() );
	
	switch (m_IndexType) {
		case MorphXML_Index:	HitTypesStr = "[s:sentence:default]";
								break;

		case DWDS_Index:	if (!m_bQueryOnlyFiles)
								HitTypesStr = "[s:sentence:default]";
							break;

		case Free_Index:	if ( HitTypesStr.empty() )
							{
								fprintf (stderr, "Error! Cannot find a hit types definition in the options file.\n");
								return false;
							};
							break;
	};


	string ErrorStr;
	if (!m_Bibl.RegisterFreeBiblAttributes(FreeBiblDescr, ErrorStr))
	{
		ErrorMessage( Format("Cannot parse %s\n (%s)", FreeBiblDescr.c_str(), ErrorStr.c_str()));
		return false;
	}
	fprintf (stderr,"%s", m_Bibl.GetFreeBibiAttributesDescr().c_str());

	if (!m_Bibl.RegisterTextAreas(TextAreasDescr, ErrorStr))
	{
		ErrorMessage( Format("Cannot parse %s\n (%s)", TextAreasDescr.c_str(), ErrorStr.c_str()));
		return false;
	}
	fprintf (stderr,"%s", m_Bibl.GetTextAreasDescr().c_str());

	if (!RegisterBorderIndices(HitTypesStr.c_str()))
	{
		fprintf (stderr," Cannot parse %s\n", HitTypesStr.c_str());
		return false;
	};
	fprintf (stderr," HitBorders = %s\n", GetBorderIndicesString().c_str());

	for(size_t i=0; i < m_IndicesToShow.size(); i++)
	{
		if (m_IndicesToShow[i] >= m_Indices.size())
		{
			fprintf (stderr,"Error! index %i specified in IndicesToShow is out  of range\n", m_IndicesToShow[i]);
			return false;
		}
		if (!m_Indices[m_IndicesToShow[i]]->m_bUseItemStorage)
		{
			fprintf (stderr,"Error! index %i has no storage\n", m_IndicesToShow[i]);
			return false;
		}
	}

        for (size_t i=0; i<BigramDescr.size(); i++)
    {
        StringTokenizer fields(BigramDescr[i].c_str(), " \t\r\n");
        if (!fields())
        {
            fprintf (stderr,"cannot parse field Bigrams %s\n", BigramDescr[i].c_str());
            return false;
        }
        string IndexName = fields.val();
        if (!fields())
        {
            fprintf (stderr,"cannot parse field Bigrams %s\n", BigramDescr[i].c_str());
            return false;
        }
        CStringIndexSet* I = GetIndexByName(IndexName);
        if (!I)
        {
            fprintf (stderr,"index \"%s\" is not found, (Bigrams %s)\n", IndexName.c_str(),  BigramDescr[i].c_str());
            return false;
        }
        int MaxBigramWindowSize  = atoi (fields.val());
        if ( (MaxBigramWindowSize > 5) || (MaxBigramWindowSize <  2) )
        {
            fprintf (stderr,"MaxBigramWindowSize can be in 2,3,4 or 5\n");
            return false;
        }
        if (!fields())
        {
            fprintf (stderr,"cannot parse field Bigrams %s\n", BigramDescr[i].c_str());
            return false;
        }
        string BreakCollName = fields.val();
        int BreakCollectionNo = GetBreakCollectionIndexByName(BreakCollName);
        if (BreakCollectionNo == -1)
        {
            fprintf (stderr,"break  collection \"%s\" is not found, (Bigrams %s)\n", BreakCollName.c_str(), BigramDescr[i].c_str());
            return false;
        }

        I->m_MaxBigramWindowSize = MaxBigramWindowSize;
        I->m_BigramBorder = BreakCollectionNo;
    };



	return true;

};


string CConcIndexator::SaveOptionsToString()  const
{
	string Result;
	if (!m_LocalPathPrefix.empty())
		Result += Format("LocalPathPrefix %s\n", m_LocalPathPrefix.c_str());

	if (!m_InternetPathPrefix.empty())
		Result += Format("InternetPathPrefix %s\n", m_InternetPathPrefix.c_str());

	if (m_Language != morphUnknown)
		Result += GetStringByLanguage(m_Language)+"\n";


	if (!m_bCaseSensitive)
		Result += Format("CaseInsensitive\n");

	if (m_bShowNumberOfRelevantDocuments)
		Result += Format("ShowNumberOfRelevantDocuments\n");
	

	if (m_bUseParagraphTagToDivide)
		Result += Format("UseParagraphTagToDivide\n");
				
	if (!m_bEmptyLineIsSentenceDelim)
		Result += Format("EmptyLineIsNotSentenceDelim\n");

				
	if (!m_bUseIndention)
		Result += Format("DontUseIndention\n");
			
	if (m_bUseDwdsThesaurus)
		Result += Format("UseDwdsThesaurus\n");
			
	if (m_bIndexPunctuation)
		Result += Format("IndexPunctuation\n");

	if (m_bResumeOnIndexErrors)
		Result += Format("ResumeOnIndexErrors\n");
	

	if (m_bQueryOnlyFiles)
		Result += Format("QueryOnlyFiles\n");

	if (m_bDisableDefaultQueryLexicalExpansion)
		Result += Format("DisableDefaultQueryLexicalExpansion\n");

	if (m_LeftKwicContextSize != DefaultKwicContextSize)
		Result += Format("LeftKwicContextSize %i\n", m_LeftKwicContextSize);

	if (m_NumberOfKwicLinesInSnippets != 10)
		Result += Format("NumberOfKwicLinesInSnippets %i\n", m_NumberOfKwicLinesInSnippets);

	if (m_TfIdfRank != 0.3)
		Result += Format("TfIdfRank %f\n", m_TfIdfRank);
	if (m_NearRank != 0.3)
		Result += Format("NearRank %f\n", m_NearRank);
	if (m_PositionRank != 0.3)
		Result += Format("PositionRank %f\n", m_PositionRank);
	if (m_InterpDelimiter != "#")
	{
		Result += Format("InterpDelimiter  %s\n", m_InterpDelimiter.c_str()  );
		
	}
	if (m_IndicesToShow.size() > 1 || (m_IndicesToShow.size() == 1 &&  m_IndicesToShow[0] != 0))
	{
		string s = "IndicesToShow ";
		for(size_t i=0; i <m_IndicesToShow.size(); i++)
			s += Format("%i ",m_IndicesToShow[i]+1);
		Result += s+"\n";
	}

	if (m_RightKwicContextSize != DefaultKwicContextSize)
		Result += Format("RightKwicContextSize %i\n", m_RightKwicContextSize);

	if (m_bArchiveIndex)
		Result += Format("ArchiveIndex\n");
				
	if (m_bIndexMorphPatterns)
		Result += Format("IndexMorphPatterns\n");

	if (m_HtmlHighlighting.m_bWasReadFromString)
		Result += Format("HtmlHighlighting %s\n", m_HtmlHighlighting.ToString().c_str());

	if (m_TextHighlighting.m_bWasReadFromString)
		Result += Format("HtmlHighlighting %s\n", m_TextHighlighting.ToString().c_str());


	Result += Format("IndexType %s\n", GetIndexTypeStr() );

	if (m_bOutputBibliographyOfHits)
		Result += Format("OutputBibliographyOfHits\n");

	if (m_bIndexChunks)
		Result += Format("IndexChunks\n");

	if (m_bUserMaxTokenCountInOnePeriod)
		Result += Format("UserMaxTokenCountInOnePeriod %i\n", m_UserMaxTokenCountInOnePeriod);			

	Result += Format("MaxRegExpExpansionSize %i\n", m_MaxRegExpExpansionSize);			

	if (m_bDwdsCorpusInterface)
		Result += Format("DwdsCorpusInterface\n");

    for (size_t i=0; i < m_Indices.size(); i++)
        if  ( m_Indices[i]->UseBigrams() )
            Result += Format ("Bigrams %s %i %s\n", m_Indices[i]->m_Name.c_str(), m_Indices[i]->m_MaxBigramWindowSize, GetBreakCollectionShortName(m_Indices[i]->m_BigramBorder).c_str());

	if (m_bGutenbergInterface)
		Result += Format("GutenbergInterface\n");
	

	if (m_bNoContextOperator)
		Result += Format("NoContextOperator\n");

	string IndicesStr = GetIndicesString();
	if (!IndicesStr.empty())
		Result += "Indices " + IndicesStr + "\n";

	IndicesStr = GetBorderIndicesString();
	if (!IndicesStr.empty())
		Result += "HitBorders " + IndicesStr + "\n";
	
	Result += m_Bibl.GetFreeBibiAttributesDescr();


	return Result;
};

bool CConcIndexator::SaveOptions(string FileName) const
{
	string PathFile = MakeFName(FileName, "opt");
	FILE* fp = fopen(PathFile.c_str(), "w");
	if (!fp) return false;
	string s  = SaveOptionsToString();
	if (!s.empty())
		fprintf (fp,"%s", s.c_str());
	fclose(fp);
	return true;
};


bool	CConcIndexator::HasEqualOptions(const CConcIndexator&  X) const
{
	string s1  = SaveOptionsToString();
	string s2  = X.SaveOptionsToString();
	return s1 == s2;
};
